@@ -17,18 +17,20 @@
 #include <spa/monitor/device.h>
 #include <spa/param/audio/format.h>
 #include <spa/param/latency-utils.h>
+#include <spa/control/control.h>
 #include <spa/pod/filter.h>
 
 #include "alsa-seq.h"
 
 #define DEFAULT_DEVICE		"default"
 #define DEFAULT_CLOCK_NAME	"clock.system.monotonic"
+#define DEFAULT_DISABLE_LONGNAME true
 
 static void reset_props(struct props *props)
 {
 	strncpy(props->device, DEFAULT_DEVICE, sizeof(props->device));
 	strncpy(props->clock_name, DEFAULT_CLOCK_NAME, sizeof(props->clock_name));
-	props->disable_longname = 0;
+	props->disable_longname = DEFAULT_DISABLE_LONGNAME;
 }
 
 static int impl_node_enum_params(void *object, int seq,
@@ -225,15 +227,18 @@ static void emit_port_info(struct seq_state *this, struct seq_port *port, bool f
 	if (full)
 		port->info.change_mask = port->info_all;
 	if (port->info.change_mask) {
-		struct spa_dict_item items[5];
+		struct spa_dict_item items[6];
 		uint32_t n_items = 0;
-		int id;
+		int card_id;
 		snd_seq_port_info_t *info;
 		snd_seq_client_info_t *client_info;
+		const char *client_name, *port_name, *dir, *pn;
+		char prefix[32] = "";
 		char card[8];
 		char name[256];
 		char path[128];
 		char alias[128];
+		char stream[32];
 
 		snd_seq_port_info_alloca(&info);
 		snd_seq_get_any_port_info(this->sys.hndl,
@@ -243,54 +248,40 @@ static void emit_port_info(struct seq_state *this, struct seq_port *port, bool f
 		snd_seq_get_any_client_info(this->sys.hndl,
 				port->addr.client, client_info);
 
-		int card_id;
+		card_id = snd_seq_client_info_get_card(client_info);
+		client_name = snd_seq_client_info_get_name(client_info);
+		port_name = snd_seq_port_info_get_name(info);
+		dir = port->direction == SPA_DIRECTION_OUTPUT ? "capture" : "playback";
 
-		// Failed to obtain card number (software device) or disabled
-		if (this->props.disable_longname || (card_id = snd_seq_client_info_get_card(client_info)) < 0) {
-			snprintf(name, sizeof(name), "%s:(%s_%d) %s",
-					snd_seq_client_info_get_name(client_info),
-					port->direction == SPA_DIRECTION_OUTPUT ? "capture" : "playback",
-					port->addr.port,
-					snd_seq_port_info_get_name(info));
-		} else {
-			char *longname;
-			if (snd_card_get_longname(card_id, &longname) == 0) {
-				snprintf(name, sizeof(name), "%s:(%s_%d) %s",
-						longname,
-						port->direction == SPA_DIRECTION_OUTPUT ? "capture" : "playback",
-						port->addr.port,
-						snd_seq_port_info_get_name(info));
-				free(longname);
-			} else {
-				// At least add card number to be distinct
-				snprintf(name, sizeof(name), "%s %d:(%s_%d) %s",
-						snd_seq_client_info_get_name(client_info),
-						card_id,
-						port->direction == SPA_DIRECTION_OUTPUT ? "capture" : "playback",
-						port->addr.port,
-						snd_seq_port_info_get_name(info));
-			}
-		}
+		if (!this->props.disable_longname)
+			snprintf(prefix, sizeof(prefix), "[%d:%d] ",
+					port->addr.client, port->addr.port);
+
+		pn = port_name;
+		if (spa_strstartswith(pn, client_name))
+			pn += strlen(client_name);
+
+		snprintf(name, sizeof(name), "%s%s%s (%s)", prefix,
+				client_name, pn, dir);
 		clean_name(name);
 
-		snprintf(path, sizeof(path), "alsa:seq:%s:client_%d:%s_%d",
-				this->props.device,
-				port->addr.client,
-				port->direction == SPA_DIRECTION_OUTPUT ? "capture" : "playback",
-				port->addr.port);
+		snprintf(stream, sizeof(stream), "client_%d", port->addr.client);
+		clean_name(stream);
+
+		snprintf(path, sizeof(path), "alsa:seq:%s:%s:%s_%d",
+				this->props.device, stream, dir, port->addr.port);
 		clean_name(path);
 
-		snprintf(alias, sizeof(alias), "%s:%s",
-				snd_seq_client_info_get_name(client_info),
-				snd_seq_port_info_get_name(info));
+		snprintf(alias, sizeof(alias), "%s:%s", client_name, port_name);
 		clean_name(alias);
 
-		items[n_items++] = SPA_DICT_ITEM_INIT(SPA_KEY_FORMAT_DSP, "8 bit raw midi");
+		items[n_items++] = SPA_DICT_ITEM_INIT(SPA_KEY_FORMAT_DSP, "32 bit raw UMP");
 		items[n_items++] = SPA_DICT_ITEM_INIT(SPA_KEY_OBJECT_PATH, path);
 		items[n_items++] = SPA_DICT_ITEM_INIT(SPA_KEY_PORT_NAME, name);
 		items[n_items++] = SPA_DICT_ITEM_INIT(SPA_KEY_PORT_ALIAS, alias);
-		if ((id = snd_seq_client_info_get_card(client_info)) != -1) {
-			snprintf(card, sizeof(card), "%d", id);
+		items[n_items++] = SPA_DICT_ITEM_INIT(SPA_KEY_PORT_GROUP, stream);
+		if (card_id != -1) {
+			snprintf(card, sizeof(card), "%d", card_id);
 			items[n_items++] = SPA_DICT_ITEM_INIT(SPA_KEY_API_ALSA_CARD, card);
 		}
 		port->info.props = &SPA_DICT_INIT(items, n_items);
@@ -538,7 +529,8 @@ impl_node_port_enum_params(void *object, int seq,
 		param = spa_pod_builder_add_object(&b,
 			SPA_TYPE_OBJECT_Format, SPA_PARAM_EnumFormat,
 			SPA_FORMAT_mediaType,      SPA_POD_Id(SPA_MEDIA_TYPE_application),
-			SPA_FORMAT_mediaSubtype,   SPA_POD_Id(SPA_MEDIA_SUBTYPE_control));
+			SPA_FORMAT_mediaSubtype,   SPA_POD_Id(SPA_MEDIA_SUBTYPE_control),
+			SPA_FORMAT_CONTROL_types,  SPA_POD_CHOICE_FLAGS_Int(1u<<SPA_CONTROL_UMP));
 		break;
 
 	case SPA_PARAM_Format:
@@ -549,7 +541,8 @@ impl_node_port_enum_params(void *object, int seq,
 		param = spa_pod_builder_add_object(&b,
 			SPA_TYPE_OBJECT_Format, SPA_PARAM_Format,
 			SPA_FORMAT_mediaType,      SPA_POD_Id(SPA_MEDIA_TYPE_application),
-			SPA_FORMAT_mediaSubtype,   SPA_POD_Id(SPA_MEDIA_SUBTYPE_control));
+			SPA_FORMAT_mediaSubtype,   SPA_POD_Id(SPA_MEDIA_SUBTYPE_control),
+			SPA_FORMAT_CONTROL_types,  SPA_POD_Int(1u<<SPA_CONTROL_UMP));
 		break;
 
 	case SPA_PARAM_Buffers:
@@ -563,7 +556,7 @@ impl_node_port_enum_params(void *object, int seq,
 			SPA_PARAM_BUFFERS_buffers, SPA_POD_CHOICE_RANGE_Int(2, 1, MAX_BUFFERS),
 			SPA_PARAM_BUFFERS_blocks,  SPA_POD_Int(1),
 			SPA_PARAM_BUFFERS_size,    SPA_POD_CHOICE_RANGE_Int(
-							4096, 4096, INT32_MAX),
+							this->quantum_limit, this->quantum_limit, INT32_MAX),
 			SPA_PARAM_BUFFERS_stride,  SPA_POD_Int(1));
 		break;
 
@@ -642,6 +635,7 @@ static int port_set_format(void *object, struct seq_port *port,
 		port->have_format = false;
 	} else {
 		struct spa_audio_info info = { 0 };
+		uint32_t types;
 
 		if ((err = spa_format_parse(format, &info.media_type, &info.media_subtype)) < 0)
 			return err;
@@ -650,6 +644,13 @@ static int port_set_format(void *object, struct seq_port *port,
 		    info.media_subtype != SPA_MEDIA_SUBTYPE_control)
 			return -EINVAL;
 
+		if ((err = spa_pod_parse_object(format,
+				SPA_TYPE_OBJECT_Format, NULL,
+				SPA_FORMAT_CONTROL_types,  SPA_POD_Int(&types))) < 0)
+			return err;
+		if (types != 1u << SPA_CONTROL_UMP)
+			return -EINVAL;
+
 		port->current_format = info;
 		port->have_format = true;
 	}
@@ -927,6 +928,10 @@ impl_init(const struct spa_handle_factory *factory,
 	this->info.n_params = N_NODE_PARAMS;
 	reset_props(&this->props);
 
+	this->quantum_limit = 8192;
+	this->min_pool_size = 500;
+	this->max_pool_size = 2000;
+
 	for (i = 0; info && i < info->n_items; i++) {
 		const char *k = info->items[i].key;
 		const char *s = info->items[i].value;
@@ -936,8 +941,14 @@ impl_init(const struct spa_handle_factory *factory,
 		} else if (spa_streq(k, "clock.name")) {
 			spa_scnprintf(this->props.clock_name,
 					sizeof(this->props.clock_name), "%s", s);
+		} else if (spa_streq(k, "clock.quantum-limit")) {
+			spa_atou32(s, &this->quantum_limit, 0);
 		} else if (spa_streq(k, SPA_KEY_API_ALSA_DISABLE_LONGNAME)) {
 			this->props.disable_longname = spa_atob(s);
+		} else if (spa_streq(k, "api.alsa.seq.min-pool")) {
+			spa_atou32(s, &this->min_pool_size, 0);
+		} else if (spa_streq(k, "api.alsa.seq.max-pool")) {
+			spa_atou32(s, &this->max_pool_size, 0);
 		}
 	}
 
@@ -974,7 +985,14 @@ impl_enum_interface_info(const struct spa_handle_factory *factory,
 static const struct spa_dict_item info_items[] = {
 	{ SPA_KEY_FACTORY_AUTHOR, "Wim Taymans <wim.taymans@gmail.com>" },
 	{ SPA_KEY_FACTORY_DESCRIPTION, "Bridge midi ports with the alsa sequencer API" },
-	{ SPA_KEY_FACTORY_USAGE, "["SPA_KEY_API_ALSA_PATH"=<device>]" },
+	{ SPA_KEY_FACTORY_USAGE,
+		"["SPA_KEY_API_ALSA_PATH"=<device, default \"default\">] "
+		"[ clock.name=<clock name, default \"clock.system.monotonic\">] "
+		"[ clock.quantum-limit=<limit, default 8192>] "
+		"["SPA_KEY_API_ALSA_DISABLE_LONGNAME"=<bool, default false>] "
+		"[ api.alsa.seq.min-pool=<min-pool, default 500>] "
+		"[ api.alsa.seq.max-pool=<max-pool, default 2000>]"
+	},
 };
 
 static const struct spa_dict info = SPA_DICT_INIT_ARRAY(info_items);
