@@ -38,6 +38,7 @@ GST_DEBUG_CATEGORY_STATIC (pipewire_sink_debug);
 
 #define DEFAULT_PROP_MODE GST_PIPEWIRE_SINK_MODE_DEFAULT
 #define DEFAULT_PROP_SLAVE_METHOD GST_PIPEWIRE_SINK_SLAVE_METHOD_NONE
+#define DEFAULT_PROP_USE_BUFFERPOOL USE_BUFFERPOOL_AUTO
 
 #define MIN_BUFFERS     8u
 
@@ -51,7 +52,8 @@ enum
   PROP_STREAM_PROPERTIES,
   PROP_MODE,
   PROP_FD,
-  PROP_SLAVE_METHOD
+  PROP_SLAVE_METHOD,
+  PROP_USE_BUFFERPOOL,
 };
 
 GType
@@ -120,6 +122,8 @@ static GstCaps *gst_pipewire_sink_sink_fixate (GstBaseSink * bsink,
 static GstFlowReturn gst_pipewire_sink_render (GstBaseSink * psink,
     GstBuffer * buffer);
 
+static  gboolean gst_pipewire_sink_event (GstBaseSink *sink, GstEvent *event);
+
 static GstClock *
 gst_pipewire_sink_provide_clock (GstElement * elem)
 {
@@ -162,7 +166,9 @@ gst_pipewire_sink_propose_allocation (GstBaseSink * bsink, GstQuery * query)
 {
   GstPipeWireSink *pwsink = GST_PIPEWIRE_SINK (bsink);
 
-  gst_query_add_allocation_pool (query, GST_BUFFER_POOL_CAST (pwsink->stream->pool), 0, 0, 0);
+  if (pwsink->use_bufferpool != USE_BUFFERPOOL_NO)
+    gst_query_add_allocation_pool (query, GST_BUFFER_POOL_CAST (pwsink->stream->pool), 0, 0, 0);
+
   gst_query_add_allocation_meta (query, GST_VIDEO_META_API_TYPE, NULL);
   return TRUE;
 }
@@ -257,6 +263,14 @@ gst_pipewire_sink_class_init (GstPipeWireSinkClass * klass)
                                                       G_PARAM_READWRITE |
                                                       G_PARAM_STATIC_STRINGS));
 
+  g_object_class_install_property (gobject_class,
+                                   PROP_USE_BUFFERPOOL,
+                                   g_param_spec_boolean ("use-bufferpool",
+                                                      "Use bufferpool",
+                                                      "Use bufferpool (default: true for video, false for audio)",
+                                                      DEFAULT_PROP_USE_BUFFERPOOL,
+                                                      G_PARAM_READWRITE |
+                                                      G_PARAM_STATIC_STRINGS));
 
   gstelement_class->provide_clock = gst_pipewire_sink_provide_clock;
   gstelement_class->change_state = gst_pipewire_sink_change_state;
@@ -272,6 +286,7 @@ gst_pipewire_sink_class_init (GstPipeWireSinkClass * klass)
   gstbasesink_class->fixate = gst_pipewire_sink_sink_fixate;
   gstbasesink_class->propose_allocation = gst_pipewire_sink_propose_allocation;
   gstbasesink_class->render = gst_pipewire_sink_render;
+  gstbasesink_class->event = gst_pipewire_sink_event;
 
   GST_DEBUG_CATEGORY_INIT (pipewire_sink_debug, "pipewiresink", 0,
       "PipeWire Sink");
@@ -340,6 +355,8 @@ gst_pipewire_sink_init (GstPipeWireSink * sink)
   sink->stream =  gst_pipewire_stream_new (GST_ELEMENT (sink));
 
   sink->mode = DEFAULT_PROP_MODE;
+  sink->use_bufferpool = DEFAULT_PROP_USE_BUFFERPOOL;
+  sink->is_video = false;
 
   GST_OBJECT_FLAG_SET (sink, GST_ELEMENT_FLAG_PROVIDE_CLOCK);
 
@@ -350,12 +367,14 @@ static GstCaps *
 gst_pipewire_sink_sink_fixate (GstBaseSink * bsink, GstCaps * caps)
 {
   GstStructure *structure;
+  GstPipeWireSink *pwsink = GST_PIPEWIRE_SINK(bsink);
 
   caps = gst_caps_make_writable (caps);
 
   structure = gst_caps_get_structure (caps, 0);
 
   if (gst_structure_has_name (structure, "video/x-raw")) {
+    pwsink->is_video = true;
     gst_structure_fixate_field_nearest_int (structure, "width", 320);
     gst_structure_fixate_field_nearest_int (structure, "height", 240);
     gst_structure_fixate_field_nearest_fraction (structure, "framerate", 30, 1);
@@ -445,6 +464,13 @@ gst_pipewire_sink_set_property (GObject * object, guint prop_id,
       pwsink->slave_method = g_value_get_enum (value);
       break;
 
+    case PROP_USE_BUFFERPOOL:
+      if(g_value_get_boolean (value))
+        pwsink->use_bufferpool = USE_BUFFERPOOL_YES;
+      else
+        pwsink->use_bufferpool = USE_BUFFERPOOL_NO;
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -490,6 +516,10 @@ gst_pipewire_sink_get_property (GObject * object, guint prop_id,
       g_value_set_enum (value, pwsink->slave_method);
       break;
 
+    case PROP_USE_BUFFERPOOL:
+      g_value_set_boolean (value, !!pwsink->use_bufferpool);
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -543,10 +573,7 @@ static void rate_match_resample(GstPipeWireSink *pwsink)
 		    stream->err_avg, stream->err_var, stream->dll.bw);
   }
 
-  if (pwsink->match) {
-    pwsink->match->rate = corr;
-    SPA_FLAG_UPDATE(pwsink->match->flags, SPA_IO_RATE_MATCH_FLAG_ACTIVE, true);
-  }
+  pw_stream_set_rate (stream->pwstream, corr);
 }
 
 static void
@@ -630,7 +657,10 @@ do_send_buffer (GstPipeWireSink *pwsink, GstBuffer *buffer)
   }
 
   if ((res = pw_stream_queue_buffer (stream->pwstream, data->b)) < 0) {
-    g_warning ("can't send buffer %s", spa_strerror(res));
+    GST_WARNING_OBJECT (pwsink, "can't send buffer %s", spa_strerror(res));
+  } else {
+    data->queued = TRUE;
+    GST_LOG_OBJECT(pwsink, "queued pwbuffer: %p; gstbuffer %p ",data->b, buffer);
   }
 
   switch (pwsink->slave_method) {
@@ -681,18 +711,6 @@ on_state_changed (void *data, enum pw_stream_state old, enum pw_stream_state sta
   pw_thread_loop_signal (pwsink->stream->core->loop, FALSE);
 }
 
-static void
-on_io_changed (void *obj, uint32_t id, void *data, uint32_t size)
-{
-  GstPipeWireSink *pwsink = obj;
-
-  switch (id) {
-    case SPA_IO_RateMatch:
-      pwsink->match = data;
-      break;
-  }
-}
-
 static void
 on_param_changed (void *data, uint32_t id, const struct spa_pod *param)
 {
@@ -719,7 +737,6 @@ gst_pipewire_sink_setcaps (GstBaseSink * bsink, GstCaps * caps)
   g_autoptr(GPtrArray) possible = NULL;
   enum pw_stream_state state;
   const char *error = NULL;
-  gboolean res = FALSE;
   GstStructure *config, *s;
   guint size;
   guint min_buffers;
@@ -734,9 +751,14 @@ gst_pipewire_sink_setcaps (GstBaseSink * bsink, GstCaps * caps)
     gst_structure_get_int (s, "rate", &rate);
     pwsink->rate = rate;
     pwsink->rate_match = true;
+
+    /* Don't provide bufferpool for audio if not requested by the application/user */
+    if (pwsink->use_bufferpool != USE_BUFFERPOOL_YES)
+      pwsink->use_bufferpool = USE_BUFFERPOOL_NO;
   } else {
     pwsink->rate = rate = 0;
     pwsink->rate_match = false;
+    pwsink->is_video = true;
   }
 
   spa_dll_set_bw(&pwsink->stream->dll, SPA_DLL_BW_MIN, 4096, rate);
@@ -810,20 +832,21 @@ gst_pipewire_sink_setcaps (GstBaseSink * bsink, GstCaps * caps)
       }
     }
   }
-  res = TRUE;
 
   gst_pipewire_clock_reset (GST_PIPEWIRE_CLOCK (pwsink->stream->clock), 0);
 
   config = gst_buffer_pool_get_config (GST_BUFFER_POOL_CAST (pwsink->stream->pool));
   gst_buffer_pool_config_get_params (config, NULL, &size, &min_buffers, &max_buffers);
   gst_buffer_pool_config_set_params (config, caps, size, min_buffers, max_buffers);
+  if(pwsink->is_video)
+    gst_buffer_pool_config_add_option(config, GST_BUFFER_POOL_OPTION_VIDEO_META);
   gst_buffer_pool_set_config (GST_BUFFER_POOL_CAST (pwsink->stream->pool), config);
 
   pw_thread_loop_unlock (pwsink->stream->core->loop);
 
-  pwsink->negotiated = res;
+  pwsink->negotiated = TRUE;
 
-  return res;
+  return TRUE;
 
 start_error:
   {
@@ -839,7 +862,6 @@ gst_pipewire_sink_render (GstBaseSink * bsink, GstBuffer * buffer)
   GstPipeWireSink *pwsink;
   GstFlowReturn res = GST_FLOW_OK;
   const char *error = NULL;
-  gboolean unref_buffer = FALSE;
 
   pwsink = GST_PIPEWIRE_SINK (bsink);
 
@@ -872,35 +894,100 @@ gst_pipewire_sink_render (GstBaseSink * bsink, GstBuffer * buffer)
     goto done_unlock;
 
   if (buffer->pool != GST_BUFFER_POOL_CAST (pwsink->stream->pool)) {
-    GstBuffer *b = NULL;
-    GstMapInfo info = { 0, };
-    GstBufferPoolAcquireParams params = { 0, };
+    gsize offset = 0;
+    gsize buf_size = gst_buffer_get_size (buffer);
 
-    pw_thread_loop_unlock (pwsink->stream->core->loop);
+    GST_TRACE_OBJECT(pwsink, "Buffer is not from pipewirepool, copying into our pool");
 
-    if ((res = gst_buffer_pool_acquire_buffer (GST_BUFFER_POOL_CAST (pwsink->stream->pool), &b, &params)) != GST_FLOW_OK)
-      goto done;
+    /* For some streams, the buffer size is changed and may exceed the acquired
+     * buffer size which is acquired from the pool of pipewiresink. Need split
+     * the buffer and send them in turn for this case */
+    while (buf_size) {
+      GstBuffer *b = NULL;
+      GstMapInfo info = { 0, };
+      GstBufferPoolAcquireParams params = { 0, };
 
-    gst_buffer_map (b, &info, GST_MAP_WRITE);
-    gst_buffer_extract (buffer, 0, info.data, info.maxsize);
-    gst_buffer_unmap (b, &info);
-    gst_buffer_resize (b, 0, gst_buffer_get_size (buffer));
-    gst_buffer_copy_into(b, buffer, GST_BUFFER_COPY_METADATA, 0, -1);
-    buffer = b;
-    unref_buffer = TRUE;
+      pw_thread_loop_unlock (pwsink->stream->core->loop);
 
-    pw_thread_loop_lock (pwsink->stream->core->loop);
-    if (pw_stream_get_state (pwsink->stream->pwstream, &error) != PW_STREAM_STATE_STREAMING)
-      goto done_unlock;
+      params.flags = GST_BUFFER_POOL_ACQUIRE_FLAG_LAST;
+      res = gst_buffer_pool_acquire_buffer (GST_BUFFER_POOL_CAST (pwsink->stream->pool),
+          &b, &params);
+      if (res == GST_FLOW_CUSTOM_ERROR_1) {
+        res = gst_base_sink_wait_preroll (bsink);
+        if (res != GST_FLOW_OK)
+          goto done;
+        continue;
+      }
+      if (res != GST_FLOW_OK)
+        goto done;
+
+      if (pwsink->is_video) {
+        GstVideoFrame src, dst;
+        gboolean copied = FALSE;
+        buf_size = 0; // to break from the loop
+
+        /*
+          splitting of buffers in the case of video might break the frame layout
+          and that seems to be causing issues while retrieving the buffers on the receiver
+          side. Hence use the video_frame_map to copy the buffer of bigger size into the
+          pipewirepool's buffer
+        */
+
+        if (!gst_video_frame_map (&dst, &pwsink->stream->pool->video_info, b,
+          GST_MAP_WRITE)) {
+          GST_ERROR_OBJECT(pwsink, "Failed to map dest buffer");
+          return GST_FLOW_ERROR;
+        }
+
+        if (!gst_video_frame_map (&src, &pwsink->stream->pool->video_info, buffer, GST_MAP_READ)) {
+          gst_video_frame_unmap (&dst);
+          GST_ERROR_OBJECT(pwsink, "Failed to map src buffer");
+          return GST_FLOW_ERROR;
+        }
+
+        copied = gst_video_frame_copy (&dst, &src);
+
+        gst_video_frame_unmap (&src);
+        gst_video_frame_unmap (&dst);
+
+        if (!copied) {
+          GST_ERROR_OBJECT(pwsink, "Failed to copy the frame");
+          return GST_FLOW_ERROR;
+        }
+
+        gst_buffer_copy_into(b, buffer, GST_BUFFER_COPY_METADATA, 0, -1);
+      } else {
+        gst_buffer_map (b, &info, GST_MAP_WRITE);
+        gsize extract_size = (buf_size <= info.maxsize) ? buf_size: info.maxsize;
+        gst_buffer_extract (buffer, offset, info.data, info.maxsize);
+        gst_buffer_unmap (b, &info);
+        gst_buffer_resize (b, 0, extract_size);
+        gst_buffer_copy_into(b, buffer, GST_BUFFER_COPY_METADATA, 0, -1);
+        buf_size -= extract_size;
+        offset += extract_size;
+      }
+
+      pw_thread_loop_lock (pwsink->stream->core->loop);
+      if (pw_stream_get_state (pwsink->stream->pwstream, &error) != PW_STREAM_STATE_STREAMING) {
+        gst_buffer_unref (b);
+        goto done_unlock;
+      }
+
+      do_send_buffer (pwsink, b);
+      gst_buffer_unref (b);
+
+      if (pw_stream_is_driving (pwsink->stream->pwstream))
+        pw_stream_trigger_process (pwsink->stream->pwstream);
+    }
+  } else {
+    GST_TRACE_OBJECT(pwsink, "Buffer is from pipewirepool");
+
+    do_send_buffer (pwsink, buffer);
+
+    if (pw_stream_is_driving (pwsink->stream->pwstream))
+      pw_stream_trigger_process (pwsink->stream->pwstream);
   }
 
-  do_send_buffer (pwsink, buffer);
-  if (unref_buffer)
-    gst_buffer_unref (buffer);
-
-  if (pw_stream_is_driving (pwsink->stream->pwstream))
-    pw_stream_trigger_process (pwsink->stream->pwstream);
-
 done_unlock:
   pw_thread_loop_unlock (pwsink->stream->core->loop);
 done:
@@ -915,7 +1002,6 @@ not_negotiated:
 static const struct pw_stream_events stream_events = {
         PW_VERSION_STREAM_EVENTS,
         .state_changed = on_state_changed,
-        .io_changed = on_io_changed,
         .param_changed = on_param_changed,
         .add_buffer = on_add_buffer,
         .remove_buffer = on_remove_buffer,
@@ -943,20 +1029,14 @@ gst_pipewire_sink_change_state (GstElement * element, GstStateChange transition)
       pw_thread_loop_lock (this->stream->core->loop);
       pw_stream_set_active(this->stream->pwstream, false);
       pw_thread_loop_unlock (this->stream->core->loop);
-      break;
-    case GST_STATE_CHANGE_PAUSED_TO_PLAYING:
-      /* uncork and start play */
-      pw_thread_loop_lock (this->stream->core->loop);
-      pw_stream_set_active(this->stream->pwstream, true);
-      pw_thread_loop_unlock (this->stream->core->loop);
-      gst_buffer_pool_set_flushing(GST_BUFFER_POOL_CAST(this->stream->pool), FALSE);
+      gst_pipewire_pool_set_paused(this->stream->pool, TRUE);
       break;
     case GST_STATE_CHANGE_PLAYING_TO_PAUSED:
       /* stop play ASAP by corking */
+      gst_pipewire_pool_set_paused(this->stream->pool, TRUE);
       pw_thread_loop_lock (this->stream->core->loop);
       pw_stream_set_active(this->stream->pwstream, false);
       pw_thread_loop_unlock (this->stream->core->loop);
-      gst_buffer_pool_set_flushing(GST_BUFFER_POOL_CAST(this->stream->pool), TRUE);
       break;
     default:
       break;
@@ -965,6 +1045,16 @@ gst_pipewire_sink_change_state (GstElement * element, GstStateChange transition)
   ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);
 
   switch (transition) {
+    case GST_STATE_CHANGE_PAUSED_TO_PLAYING:
+      /* For some cases, the param_changed event is earlier than the state switch
+       * from paused state to playing state which will wait until buffer pool is ready.
+       * Guarantee to finish preoll if needed to active buffer pool before uncorking and
+       * starting play */
+      gst_pipewire_pool_set_paused(this->stream->pool, FALSE);
+      pw_thread_loop_lock (this->stream->core->loop);
+      pw_stream_set_active(this->stream->pwstream, true);
+      pw_thread_loop_unlock (this->stream->core->loop);
+      break;
     case GST_STATE_CHANGE_PLAYING_TO_PAUSED:
       break;
     case GST_STATE_CHANGE_PAUSED_TO_READY:
@@ -985,3 +1075,42 @@ open_failed:
     return GST_STATE_CHANGE_FAILURE;
   }
 }
+
+static  gboolean gst_pipewire_sink_event (GstBaseSink *sink, GstEvent *event) {
+  GstPipeWireSink *pw_sink = GST_PIPEWIRE_SINK(sink);
+  GstState current_state = GST_ELEMENT(sink)->current_state;
+
+  switch (GST_EVENT_TYPE (event)) {
+    case GST_EVENT_FLUSH_START:
+    {
+      GST_DEBUG_OBJECT (pw_sink, "flush-start");
+      pw_thread_loop_lock (pw_sink->stream->core->loop);
+
+      /* The stream would be already inactive if the sink is not PLAYING */
+      if (current_state == GST_STATE_PLAYING)
+        pw_stream_set_active(pw_sink->stream->pwstream, false);
+
+      gst_buffer_pool_set_flushing(GST_BUFFER_POOL_CAST(pw_sink->stream->pool), TRUE);
+      pw_stream_flush(pw_sink->stream->pwstream, false);
+      pw_thread_loop_unlock (pw_sink->stream->core->loop);
+      break;
+    }
+    case GST_EVENT_FLUSH_STOP:
+    {
+      GST_DEBUG_OBJECT (pw_sink, "flush-stop");
+      pw_thread_loop_lock (pw_sink->stream->core->loop);
+
+      /* The stream needs to remain inactive if the sink is not PLAYING */
+      if (current_state == GST_STATE_PLAYING)
+        pw_stream_set_active(pw_sink->stream->pwstream, true);
+
+      gst_buffer_pool_set_flushing(GST_BUFFER_POOL_CAST(pw_sink->stream->pool), FALSE);
+      pw_thread_loop_unlock (pw_sink->stream->core->loop);
+      break;
+    }
+    default:
+      break;
+  }
+
+  return GST_BASE_SINK_CLASS (parent_class)->event (sink, event);
+}
