@@ -271,6 +271,9 @@ struct impl {
 	bool mute;
 	float volume;
 
+	struct spa_latency_info latency_info;
+	struct spa_process_latency_info process_latency;
+
 	struct spa_ringbuffer ring;
 	uint8_t buffer[BUFFER_SIZE];
 
@@ -509,21 +512,11 @@ static void stream_send_packet(void *data, struct iovec *iov, size_t iovlen)
 	out_vec[msg.msg_iovlen++] = (struct iovec) { header, 12 };
 	out_vec[msg.msg_iovlen++] = (struct iovec) { out, len };
 
-	pw_log_debug("raop sending %ld", out_vec[0].iov_len + out_vec[1].iov_len + out_vec[2].iov_len);
+	pw_log_debug("raop sending %zu", out_vec[0].iov_len + out_vec[1].iov_len + out_vec[2].iov_len);
 
 	send_packet(impl->server_fd, &msg);
 }
 
-static inline void
-set_iovec(struct spa_ringbuffer *rbuf, void *buffer, uint32_t size,
-		uint32_t offset, struct iovec *iov, uint32_t len)
-{
-	iov[0].iov_len = SPA_MIN(len, size - offset);
-	iov[0].iov_base = SPA_PTROFF(buffer, offset, void);
-	iov[1].iov_len = len - iov[0].iov_len;
-	iov[1].iov_base = buffer;
-}
-
 static int create_udp_socket(struct impl *impl, uint16_t *port)
 {
 	int res, ip_version, fd, val, i, af;
@@ -858,15 +851,29 @@ static uint32_t msec_to_samples(struct impl *impl, uint32_t msec)
 	return (uint64_t) msec * impl->rate / 1000;
 }
 
+static void update_latency(struct impl *impl)
+{
+	uint32_t n_params = 0;
+	const struct spa_pod *params[3];
+	uint8_t buffer[1024];
+	struct spa_pod_builder b;
+	struct spa_latency_info latency;
+
+	spa_pod_builder_init(&b, buffer, sizeof(buffer));
+
+	latency = SPA_LATENCY_INFO(PW_DIRECTION_INPUT);
+
+	spa_process_latency_info_add(&impl->process_latency, &latency);
+	params[n_params++] = spa_latency_build(&b, SPA_PARAM_Latency, &latency);
+	params[n_params++] = spa_latency_build(&b, SPA_PARAM_Latency, &impl->latency_info);
+	params[n_params++] = spa_process_latency_build(&b, SPA_PARAM_ProcessLatency, &impl->process_latency);
+	rtp_stream_update_params(impl->stream, params, n_params);
+}
+
 static int rtsp_record_reply(void *data, int status, const struct spa_dict *headers, const struct pw_array *content)
 {
 	struct impl *impl = data;
 	const char *str;
-	uint32_t n_params;
-	const struct spa_pod *params[2];
-	uint8_t buffer[1024];
-	struct spa_pod_builder b;
-	struct spa_latency_info latency;
 	char progress[128];
 	struct timespec timeout, interval;
 
@@ -884,25 +891,21 @@ static int rtsp_record_reply(void *data, int status, const struct spa_dict *head
 	interval.tv_sec = 2;
 	interval.tv_nsec = 0;
 
-	if (!impl->feedback_timer)
+	// feedback timer is only needed for auth_setup	encryption
+	if (impl->encryption == CRYPTO_AUTH_SETUP && !impl->feedback_timer) {
+
 		impl->feedback_timer = pw_loop_add_timer(impl->loop, rtsp_do_post_feedback, impl);
-	pw_loop_update_timer(impl->loop, impl->feedback_timer, &timeout, &interval, false);
+		pw_loop_update_timer(impl->loop, impl->feedback_timer, &timeout, &interval, false);
+	}
 
 	if ((str = spa_dict_lookup(headers, "Audio-Latency")) != NULL) {
 		uint32_t l;
 		if (spa_atou32(str, &l, 0))
 			impl->latency = SPA_MAX(l, impl->latency);
 	}
+	impl->process_latency.rate = impl->latency + msec_to_samples(impl, RAOP_LATENCY_MS);
 
-	spa_zero(latency);
-	latency.direction = PW_DIRECTION_INPUT;
-	latency.min_rate = latency.max_rate = impl->latency + msec_to_samples(impl, RAOP_LATENCY_MS);
-
-	n_params = 0;
-	spa_pod_builder_init(&b, buffer, sizeof(buffer));
-	params[n_params++] = spa_latency_build(&b, SPA_PARAM_Latency, &latency);
-
-	rtp_stream_update_params(impl->stream, params, n_params);
+	update_latency(impl);
 
 	rtp_stream_set_first(impl->stream);
 
@@ -1328,12 +1331,10 @@ static int rtsp_post_auth_setup_reply(void *data, int status, const struct spa_d
 
 static int rtsp_do_post_auth_setup(struct impl *impl)
 {
-	static const uint8_t content[33] = {
-		0x01,
-		0x59, 0x02, 0xed, 0xe9, 0x0d, 0x4e, 0xf2, 0xbd,
-		0x4c, 0xb6, 0x8a, 0x63, 0x30, 0x03, 0x82, 0x07,
-		0xa9, 0x4d, 0xbd, 0x50, 0xd8, 0xaa, 0x46, 0x5b,
-		0x5d, 0x8c, 0x01, 0x2a, 0x0c, 0x7e, 0x1d, 0x4e };
+	static const unsigned char content[33] =
+		"\x01"
+		"\x59\x02\xed\xe9\x0d\x4e\xf2\xbd\x4c\xb6\x8a\x63\x30\x03\x82\x07"
+		"\xa9\x4d\xbd\x50\xd8\xaa\x46\x5b\x5d\x8c\x01\x2a\x0c\x7e\x1d\x4e";
 
 	return pw_rtsp_client_url_send(impl->rtsp, "/auth-setup", "POST", &impl->headers->dict,
 				       "application/octet-stream", content, sizeof(content),
@@ -1666,6 +1667,32 @@ static void stream_props_changed(struct impl *impl, uint32_t id, const struct sp
 	rtp_stream_set_param(impl->stream, id, param);
 }
 
+static void param_latency_changed(struct impl *impl, const struct spa_pod *param)
+{
+	struct spa_latency_info latency;
+
+	if (param == NULL || spa_latency_parse(param, &latency) < 0)
+		return;
+	if (latency.direction == SPA_DIRECTION_OUTPUT)
+		impl->latency_info = latency;
+
+	update_latency(impl);
+}
+
+static void param_process_latency_changed(struct impl *impl, const struct spa_pod *param)
+{
+	struct spa_process_latency_info info;
+
+	if (param == NULL)
+		spa_zero(info);
+	else if (spa_process_latency_parse(param, &info) < 0)
+		return;
+	if (spa_process_latency_info_compare(&impl->process_latency, &info) == 0)
+		return;
+	impl->process_latency = info;
+	update_latency(impl);
+}
+
 static void stream_param_changed(void *data, uint32_t id, const struct spa_pod *param)
 {
 	struct impl *impl = data;
@@ -1681,6 +1708,12 @@ static void stream_param_changed(void *data, uint32_t id, const struct spa_pod *
 		if (param != NULL)
 			stream_props_changed(impl, id, param);
 		break;
+	case SPA_PARAM_Latency:
+		param_latency_changed(impl, param);
+		break;
+	case SPA_PARAM_ProcessLatency:
+		param_process_latency_changed(impl, param);
+		break;
 	default:
 		break;
 	}
@@ -1810,6 +1843,8 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	impl->context = context;
 	impl->loop = pw_context_get_main_loop(context);
 
+	impl->latency_info = SPA_LATENCY_INFO(SPA_DIRECTION_OUTPUT);
+
 	ip = pw_properties_get(props, "raop.ip");
 	port = pw_properties_get(props, "raop.port");
 	if (ip == NULL || port == NULL) {
